from __future__ import annotations


class PhenotypeFactoryError(RuntimeError):
    @classmethod
    def character_not_allowed(cls, char: str) -> PhenotypeFactoryError:
        return cls(f'Character not allowed: {char}')



from random import randint
from typing import Tuple
from alge.genotype.operator import Operator
from alge.genotype.creature import Creature
from alge.genotype.population import Population
from alge.genotype.phenotype import Phenotype


class Crossover(Operator):

    POPULATION_SIZE = 'population_size'
    CHILDREN_PER_PARENTS = 'children_per_parents'

    _config: dict = {
        POPULATION_SIZE: 10,
        CHILDREN_PER_PARENTS: 2,
    }

    def __init__(self, config: dict = None):
        if config is not None:
            self._config = config

    def do(self, population: Population) -> Population:
        next_generation = Population([])
        parent_index = 0
        while next_generation.size() < self._config[self.POPULATION_SIZE]:
            parents: Tuple[Creature, Creature] = self.choose_parents(parent_index, population)
            for i in range(0, self._config[self.CHILDREN_PER_PARENTS]):
                next_generation.add(self.child(parents))
            parent_index += 2
        return next_generation

    @staticmethod
    def child(parents: Tuple[Creature, Creature]) -> Creature:
        phenotype_len = len(parents[0].phenotype())
        # if len(parents[1].phenotype()) is not phenotype_len:
        #     raise Exception('Parents have different phenotypes lengths!')

        cross_point = randint(1, phenotype_len - 1)
        # part1 = parents[0].phenotype()._data[0:cross_point]
        # part2 = parents[1].phenotype()._data[cross_point:]
        child_phenotype = Phenotype(
            parents[0].phenotype()._data[0:cross_point] + parents[1].phenotype()._data[cross_point:]
        )

        return Creature(child_phenotype)

    @staticmethod
    def choose_parents(parent_index: int, population: Population) -> Tuple[Creature, Creature]:
        parent1 = population[parent_index]
        parent2 = population[parent_index + 1]
        return parent1, parent2



import pytest

from alge.genotype.creature import Creature
from alge.genotype.operatord.crossover import Crossover
from alge.genotype.operatord.selection import Selection
from alge.genotype.phenotype import Phenotype
from alge.genotype.population import Population


@pytest.fixture
def population():
    result = Population([])
    result.add(Creature(Phenotype([False, False])))
    result.add(Creature(Phenotype([True, True])))
    return result


def test_crossover(population):
    crossover = Crossover({
        Crossover.POPULATION_SIZE: 1,
        Crossover.CHILDREN_PER_PARENTS: 1,
    })
    result = crossover.do(population)

    assert result.size() == 1
    assert result._population[0].phenotype()._data == [False, True]


def test_crossover_multi(population):
    crossover = Crossover({
        Crossover.POPULATION_SIZE: 2,
        Crossover.CHILDREN_PER_PARENTS: 2,
    })
    result = crossover.do(population)

    assert result.size() == 2
    assert result._population[0].phenotype()._data == [False, True]
    assert result._population[1].phenotype()._data == [False, True]




from alge.genotype.creature import Creature
from typing import List


class Population:

    _population: List[Creature]
    _iter_next: int = 0

    def __init__(self, population: List[Creature]):
        self._population = population

    def __getitem__(self, item: int) -> Creature:
        return self._population[item]

    def __iter__(self):
        self._iter_next = 0
        return self

    def __next__(self):
        if self._iter_next < len(self._population):
            return self._population[self._iter_next]
        else:
            raise StopIteration

    def size(self) -> int:
        return len(self._population)

    def add(self, creature: Creature) -> None:
        self._population.append(creature)

    def sort_by_fitness(self) -> None:
        self._population.sort(key=lambda item: item._fitness, reverse=True)



from ...genotype.creature import Creature
from ...genotype.phenotyped.factory import  Factory as PhenotypeFactory


def test_creature_create():
    creature = Creature(PhenotypeFactory.create_from_string('1010101'))
    assert isinstance(creature, Creature)


def test_fitness():
    creature = Creature(PhenotypeFactory.create_from_string('1010101'), 10.10)
    assert creature.fitness() == 10.10
    creature.set_fitness(0.9)
    assert creature.fitness() == 0.9




from typing import List, Optional
from logging import Logger
from .config import Config
from ..genotype.operator import Operator
from ..genotype.population import Population
from ..genotype.fitness_function import FitnessFunction
# from ..genom.operator import Operator
from alge.genalg.stats import Stats


class GeneticAlgorithmApi():

    # _logger: Logger
    # _config: Config
    # _population: Population = None
    # _fitness_function: FitnessFunction
    # _actions: List[Action]

    def __init__(self,
                 logger: Optional[Logger],
                 config: Config,
                 fitness_function: FitnessFunction,
                 actions: List[Operator],
                 stats: Stats = None
                 ):
        self._logger = logger
        self._config = config
        self._fitness_function = fitness_function
        self._actions: List[Operator] = actions
        self._population: Population = Population([])
        self._stats: Stats = stats

    def init_population(self, population: Population) -> None:
        self._population = population
        self._fit_population()

    def population(self) -> Population:
        return self._population

    def next_generation(self) -> Population:
        if self._population is None:
            raise Exception('Population not initialized!')

        self._execute_actions()
        self._fit_population()
        return self._population

    def _fit_population(self) -> None:
        for creature in self._population._population:
            creature._fitness = self._fitness_function.fit(creature.phenotype())
            # if self._stats:
            #     self._stats.add_fit_stats({})

    def _execute_actions(self) -> None:
        # next_generation = self._selection(self._population)
        # next_generation = self._crossover(next_generation)
        # next_generation = self._mutation(next_generation)

        generation = self._population
        for action in self._actions:
            generation = action.do(generation)
            if self._stats:
                self._stats.add_action_stats(action.get_stats())

        self._population = generation

    # # actions
    # def _selection(self, population: Population) -> Population:
    #     return population
    #
    # def _crossover(self, population: Population) -> Population:
    #     return population
    #
    # def _mutation(self, population: Population) -> Population:
    #     return population



import pytest

from alge.genotype.creature import Creature
from alge.genotype.operatord.selection import Selection
from alge.genotype.phenotype import Phenotype
from alge.genotype.population import Population


@pytest.fixture
def population():
    result = Population([])
    result.add(Creature(Phenotype([False, True]), 20))
    result.add(Creature(Phenotype([True, True]), 10))
    result.add(Creature(Phenotype([True, False]), 40))
    result.add(Creature(Phenotype([False, False]), 30))
    return result


def test_selection(population):
    selection = Selection()
    result = selection.do(population)

    assert result.size() == 2
    assert result._population[0].fitness() == 40
    assert result._population[1].fitness() == 30


def test_selection_custom_elements_number(population):
    selection = Selection({Selection.CREATURES_TO_CHOOSE: 3})
    result = selection.do(population)

    assert result.size() == 3
    assert result._population[0].fitness() == 40
    assert result._population[1].fitness() == 30
    assert result._population[2].fitness() == 20




from alge.genotype.phenotype import Phenotype


class Creature():

    _phenotype: Phenotype
    _fitness: float

    def __init__(self, phenotype: Phenotype, fitness: float = 0):
        self._phenotype = phenotype
        self._fitness = fitness

    def phenotype(self) -> Phenotype:
        return self._phenotype

    def fitness(self) -> float:
        return self._fitness

    def set_fitness(self, value: float) -> None:
        self._fitness = value




from typing import List
from alge.genotype.operator import Operator
from alge.genotype.creature import Creature
from alge.genotype.population import Population


class Selection(Operator):

    CREATURES_TO_CHOOSE = 'pair_to_choose'

    _config: dict = {
        CREATURES_TO_CHOOSE: 2,
    }

    def __init__(self, config: dict = None):
        if config is not None:
            self._config = config

    def do(self, population: Population) -> Population:
        parents: List[Creature] = []
        quantity: int = self._config[self.CREATURES_TO_CHOOSE]

        if quantity > population.size():
            raise Exception('Population too small to choose %d parents' % (quantity,))

        population.sort_by_fitness()

        for i in range(0, quantity):
            parents.append(population[i])

        return Population(parents)



from alge.genotype.phenotype import Phenotype
from alge.genotype.phenotyped.error.phenotype_factory_error import PhenotypeFactoryError


CHAR_0 = '0'
CHAR_1 = '1'


class Factory:

    @staticmethod
    def create_from_string(input: str) -> Phenotype:
        phenotype = []
        for char in input:
            if char == CHAR_0:
                phenotype.append(False)
            elif char == CHAR_1:
                phenotype.append(True)
            else:
                raise PhenotypeFactoryError(char)
        return Phenotype(phenotype)



from abc import ABC, abstractmethod


class Stats(ABC):

    # @abstractmethod
    # def add_fit_stats(self, stats: dict) -> None:
    #     pass

    @abstractmethod
    def add_action_stats(self, stats: dict) -> None:
        pass




from collections.abc import Sized
from typing import List


class Phenotype(Sized):

    _data: List[bool]

    def __init__(self, phenotype: List[bool]):
        self._data = phenotype

    def __len__(self) -> int:
        return len(self._data)





import pytest
from unittest import mock
from ..api import GeneticAlgorithmApi
from ..config import Config


@pytest.fixture()
def api():
    fitness_function = mock.Mock()
    return GeneticAlgorithmApi(
        None,
        Config(),
        fitness_function,
        []
    )


def test_api_instantiate(api):
    assert isinstance(api, GeneticAlgorithmApi)


def test_ng_without_init(api: GeneticAlgorithmApi):
    assert api.next_generation() == api.population()



import pytest
from typing import Tuple
from alge.genotype.phenotype import Phenotype
from alge.genotype.phenotyped.error.phenotype_factory_error import PhenotypeFactoryError
from alge.genotype.phenotyped.factory import  Factory as PhenotypeFactory


def phenotypes_to_create() -> Tuple:
    data = [
        ('',),
        ('0',),
        ('1',),
        ('00000000',),
        ('10101010001111010001101000101111100101010101010100001110001010001',),
    ]
    for row in data:
        yield row


def wrong_phenotypes() -> Tuple:
    data = [
        (' ',),
        ('a',),
        ('0110 ',),
        ('11x',),
    ]
    for row in data:
        yield row


@pytest.mark.parametrize(['phenotype_as_string'], phenotypes_to_create())
def test_phenotype_from_string_creation(phenotype_as_string: str):
    phenotype = PhenotypeFactory.create_from_string(phenotype_as_string)
    assert isinstance(phenotype, Phenotype)
    assert len(phenotype_as_string) == len(phenotype._data)


@pytest.mark.parametrize(['phenotype_as_string'], wrong_phenotypes())
def test_phenotype_from_string_error(phenotype_as_string: str):
    with pytest.raises(PhenotypeFactoryError):
        PhenotypeFactory.create_from_string(phenotype_as_string)




from abc import ABC
from abc import abstractmethod
from alge.genotype.phenotype import Phenotype


class FitnessFunction(ABC):

    @abstractmethod
    def fit(self, phenotype: Phenotype) -> float:
        pass




from abc import ABC, abstractmethod
from alge.genotype.population import Population


class Operator(ABC):

    @abstractmethod
    def do(self, population: Population) -> Population:
        pass

    def get_stats(self) -> dict:
        pass
